/* tilemap.js
Copyright © 2016 Mark Damon Hughes

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

----

Tile art is public domain from David Gervais <dgerv@sympatico.ca>

----

Usage:

<head>
<script src="tilemap.js" type="text/javascript" language="javascript"></script>
</head>
<body>

<div id="haven-area" class="tilemap" title="Haven Area" tilechars="2" scale="10km" reference="hyperborea,2,3">
<pre>
t t t t t t t t t t
t t6t . . h h h h t
t t . . t t h h h h
t . . h t . . h h h
. . h h . .2.━h━h━h━
. . . h . . . h h h
. . . . h h h h h h
t . . t t h h h M h
t t . . t t h M M M
t t . . t1h h h M M
</pre>
<dl>
<dt>1,1</dt>	<dd>Goblin Hive</dd>
<dt>5,4</dt>	<dd>Haven</dd>
<dt>4,9</dt>	<dd>Felder's Green</dd>
</dl>
</div>

… to end of file …
<script type="text/javascript" language="javascript">tileMapInit();</script>
</body>

----

- id is the internal name of the map.
- class must be "tilemap".
- title is displayed as a header 3.
- scale is displayed as a subtitle.
- tilechars is the number of characters per tile, default 2. tilechars="1" means no overlay layers, just base terrain; tilechars="2" or higher creates (tilechars-1) overlay layers.
- reference is the id,x,y of another map in the same document, you can leave that out if not used.
- pre block is the map, 'tilechars' characters per tile, a base terrain and 'tilechars'-1 other features on top, or spaces for none.
- dl group is map labels, dt must be the coordinate, "x,y" starting at 0,0, dd is the title. You can use a dash - to split long words.

- The dictionary of characters to tile filenames is in kTileNames below. The images are in img/tilemap, and must be 32x32 png images.

- To type the rivers and roads, here's a trick on the Mac: Open character viewer (Emoji & Symbols), gear icon, customize list, and check Box Drawing. Rivers are the double-lines, roads are heavy lines.

*/

var kTileNames = {
	// base terrains, use in tile 1 only
	",": "moor",
	";": "moor_tree",
	".": "grass",
	"^": "mountain",
	"+": "forest",
	"~": "ocean",
	"_": "floor",
	"#": "wall_stone",
	"%": "wall_wood",
	"=": "door_open",
	"≠": "door_closed",
	"≈": "door_secret",
	"‡": "door_cell",
	"∏": "door_wood",
	"[": "stairs_up",
	"]": "stairs_down",

	"b": "badlands",
	"c": "moor_cold",
	"d": "desert",
	"f": "field",
	"h": "hill",
	"i": "ice",
	"k": "desert_palm",
	"p": "hill_frost",
	"s": "island",
	"t": "taiga",
	"v": "volcwaste",

	"C": "coldmtn",
	"M": "stonemtn",
	"V": "volcano",

	// overlays, use in tiles 2+ only
	" ": "empty",
	"-": "window",
	"?": "sign",
	"|": "pillar",
	"F": "fountain",
	"P": "pit",
	"R": "tombstone2",
	"S": "statue",
	"T": "trapdoor",
	"W": "web2",
	"$": "chest",

	"0": "town_adobe",
	"1": "town_hut",
	"2": "town_redhouse",
	"3": "town_tower",
	"4": "town_goldcity",
	"5": "town_blackcity",
	"6": "hole",
	"7": "mine",
	"8": "ruin",
	"9": "dungeon",

	"║": "river_ns2",
	"═": "river_ew2",
	"╔": "river_es2",
	"╦": "river_esw2",
	"╗": "river_sw2",
	"╠": "river_nes2",
	"╬": "river_nesw2",
	"╣": "river_nsw2",
	"╚": "river_ne2",
	"╩": "river_new2",
	"╝": "river_nw2",

	"┃": "road_ns2",
	"━": "road_ew2",
	"┏": "road_es2",
	"┳": "road_esw2",
	"┓": "road_sw2",
	"┣": "road_nes2",
	"╋": "road_nesw2",
	"┫": "road_nsw2",
	"┗": "road_ne2",
	"┻": "road_new2",
	"┛": "road_nw2",

	"╪": "bridge_ns2",
	"╫": "bridge_ew2",

	"\x00": "error",
};

var kTileMapRedrawTimeMS = 100;
var kTileMapSize = 32;
var TileMapImages = {};
var TileMapData = {};

function isEmpty(x) {
	return x === undefined || x === null || x === 0 || x === "" || x.length === 0;
}

function readTileMap(mapElem) {
	var title = mapElem.getAttribute("title");
	var scale = mapElem.getAttribute("scale");
	var reference = mapElem.getAttribute("reference") || "";
	var tilechars = mapElem.getAttribute("tilechars");

	var preGroup = mapElem.getElementsByTagName("PRE");
	if (isEmpty(preGroup)) {
		console.error("Missing <pre> in map "+title);
		return;
	}
	var mapData = preGroup[0].innerHTML.trim().split(/[\r\n]+/);

	var labels = {};
	var labelGroup = mapElem.getElementsByTagName("DL");
	if (isEmpty(labelGroup)) {
		console.error("Missing <dl> in map "+title);
		return;
	}
	for (var i = 0, dt = null; i < labelGroup[0].children.length; ++i) {
		var elem = labelGroup[0].children[i];
		if (elem.nodeName === "DT") {
			dt = elem.innerHTML.trim();
		} else if (elem.nodeName === "DD") {
			if (dt === null) {
				console.error("Missing location for label "+elem.innerHTML);
				continue;
			}
			labels[dt] = elem.innerHTML.trim();
			dt = null;
		}
	}

	return {
		title: title,
		scale: scale,
		tilechars: tilechars,
		reference: reference,
		tiles: mapData,
		labels: labels,
	};
}

function drawTile(ctx, ch, x, y) {
	if (ch === "") { // chars past the end are empty
		ch = " ";
	}
	var t = kTileNames[ch];
	if (t === undefined) {
		t = "error";
	}
	var img = TileMapImages[t];
	ctx.drawImage(img, x * kTileMapSize, y * kTileMapSize, kTileMapSize, kTileMapSize);
}

function drawAxisLabel(ctx, x, y) {
	if (x === 0 || y === 0) {
		var text = (y === 0) ? ""+x : ""+y;
		var xx = x * kTileMapSize + 2.0;
		var yy = y * kTileMapSize + 8.0;
		ctx.strokeStyle = "black";
		ctx.strokeText(text, xx-1.0, yy-1.0);
		ctx.strokeText(text, xx+1.0, yy+1.0);
		ctx.fillStyle = "#ccff80";
		ctx.fillText(text, xx, yy);
	}
}

function drawLabel(ctx, lbl, x, y, lineHeight) {
	var xx = x * kTileMapSize + 2.0;
	var yy = y * kTileMapSize + kTileMapSize - 2.0;

	var words = lbl.split(/[\s-]+/);
	yy -= (words.length-1) * lineHeight;
	for (var i = 0; i < words.length; ++i) {
		ctx.strokeStyle = "black";
		ctx.strokeText(words[i], xx+1.0, yy+1.0);
		ctx.fillStyle = "yellow";
		ctx.fillText(words[i], xx, yy);
		yy += lineHeight;
	}
}

function renderTileMap(mapid, showKeys) {
	var mapElem = document.getElementById(mapid);
	var mapData = TileMapData[mapid];
	var tilechars = mapData.tilechars || 2;
	var width = Math.ceil(mapData.tiles[0].length / tilechars) * kTileMapSize;
	var height = mapData.tiles.length * kTileMapSize;

	var html = "<h3>"+mapData.title+"</h3>\n";

	if ( ! isEmpty(mapData.reference)) {
		var words = mapData.reference.split(",");
		var refMap = TileMapData[words[0]];
		html += "<div><i>Detail of "+(refMap ? refMap.title : words[0])+
			" ("+words[1]+","+words[2]+")</i></div>\n";
	}

	html += "<div>";
 	html += "<i>1 square = "+mapData.scale+"</i> ";
	html += "<input type='checkbox' id='"+mapid+"-showkeys' "+
		(showKeys ? "checked=\"checked\" " : "")+
		"onchange='renderTileMap(\""+mapid+"\", "+(!showKeys)+");'><label for='"+mapid+"-showkeys'>Show Map Keys</label>\n";
	html += "</div>\n";

	html += "<canvas id='"+mapElem.id+"-canvas' width='"+width+"' height='"+height+"'></canvas>\n";
	mapElem.innerHTML = html;
	//console.log(html);

	var canvas = document.getElementById(mapid+"-canvas");
	var ctx = canvas.getContext("2d");
	ctx.mozImageSmoothingEnabled = false;
	ctx.webkitImageSmoothingEnabled = false;
	ctx.msImageSmoothingEnabled = false;
	ctx.imageSmoothingEnabled = false;

	ctx.font = '6pt "Helvetica Neue", Helvetica, sans-serif';
	ctx.textAlign = "left";
	for (var y = 0; y < mapData.tiles.length; ++y) {
		var line = mapData.tiles[y];
		for (var x = 0, w = Math.ceil(mapData.tiles[y].length / tilechars); x < w; ++x) {
			for (var i = 0; i < tilechars; ++i) {
				drawTile(ctx, line.charAt(x * tilechars + i), x, y);
			}
			drawAxisLabel(ctx, x, y);
		}
	}

	if (showKeys) {
		ctx.font = '8pt "Helvetica Neue", Helvetica, sans-serif';
		var lineHeight = 9.0;
		ctx.textAlign = "left";
		for (var y = 0; y < mapData.tiles.length; ++y) {
			var line = mapData.tiles[y];
			for (var x = 0; x < line.length; ++x) {
				var lbl = mapData.labels[x+","+y];
				if ( ! isEmpty(lbl)) {
					drawLabel(ctx, lbl, x, y, lineHeight);
				}
			}
		}
	}
}

function preloadTileMapImages() {
	var imageNames = [];
	for (var k in kTileNames) {
		imageNames.push(kTileNames[k]);
	}
	for (var i = 0; i < imageNames.length; ++i) {
		var img = new Image();
		img.title = imageNames[i];
		img.addEventListener("load", function() {
			console.log(this+" loaded "+this.src);
			TileMapImages[this.title] = this;
		}, false);
		img.src = "img/tilemap/"+imageNames[i]+".png";
		console.log("Started load of "+img.src);
	}
}

function mapLoop() {
	var imageCount = Object.keys(TileMapImages).length;
	var tileCount = Object.keys(kTileNames).length;
	if (imageCount != tileCount) {
		console.log("Waiting for images, "+imageCount+"/"+tileCount);
		setTimeout(mapLoop, kTileMapRedrawTimeMS);
		return;
	}

	var mapElements = document.getElementsByClassName("tilemap");
	if (isEmpty(mapElements)) {
		console.error("No maps found.");
		return;
	}
	for (var i = 0; i < mapElements.length; ++i) {
		var mapElem = mapElements[i];
		TileMapData[mapElem.id] = readTileMap(mapElem);
	}
	for (var i = 0; i < mapElements.length; ++i) {
		var mapElem = mapElements[i];
		renderTileMap(mapElem.id, true);
	}
}

function tileMapInit() {
	preloadTileMapImages();
	setTimeout(mapLoop, kTileMapRedrawTimeMS);
}
